
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "rif_rgb.h"

extern void	Open(),Store(),Close();

static void	Error(msg)
char		*msg;
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static double	A2,B2;

static void	XYZ(lon,lat,e,x,y,z)
double		lon,lat,e,*x,*y,*z;
{
	double	c=cos(lat),s=sin(lat),ab=sqrt(A2*c*c+B2*s*s),xy=(A2/ab+e)*c;

	*x=xy*cos(lon); *y=xy*sin(lon); *z=(B2/ab+e)*s;
}

static double	xO,yO,zO,hx,hy,hz,vx,vy,vz,dx,dy,dz;

static void	HVD(lon,lat,e,h,v,d)
double		lon,lat,e,*h,*v,*d;
{
	double	x,y,z;

	XYZ(lon,lat,e,&x,&y,&z); x-=xO;
				 y-=yO;
				 z-=zO; *h=hx*x+hy*y+hz*z;
					*v=vx*x+vy*y+vz*z;
					*d=dx*x+dy*y+dz*z;
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	double	lon1,lon2,lat1,lat2,base,step,sLon,sLat,fLon,fLat,factor,depth,
		lon,lat,hO,vO,dO,h1,v1,d1,h2,v2,d2,h3,v3,d3,h4,v4,d4,
		e1=0.0,e2=0.0,e3=0.0,e4=0.0;
	int	nLon,nLat,iLon,iLat,BPS,cLon,cLat,color=0,f1=0,f2=0;
	Cell	**dem=NULL,**R=NULL,**G=NULL,**B=NULL;

	if (argc<11 || argc>14)
	    Error("usage : ssm A B lon1 lat1 lon2 lat2\n"
		  "            (nLon nLat)|(DEM base step ratio)\n"
		  "            {color}\n"
		  "            depth STL");

	A2=atof(argv[1])*atof(argv[1]);
	B2=atof(argv[2])*atof(argv[2]);

	if ((lon1=atof(argv[3])*M_PI/180.0)>=
	    (lon2=atof(argv[5])*M_PI/180.0) ||
	    (lat1=atof(argv[4])*M_PI/180.0)>=
	    (lat2=atof(argv[6])*M_PI/180.0)) Error("bad map area.");

	if (argc==11 || argc==12) {
	    nLon=atoi(argv[7]);
	    nLat=atoi(argv[8]);
	}
	else {
	    ReadImageFile(argv[7],&nLon,&nLat,NULL,&dem,NULL);
	    base=atof(argv[8])*atof(argv[10]);
	    step=atof(argv[9])*atof(argv[10]);
	}
	if (nLon<=1 || nLat<=1) Error("bad grid number.");

	--nLon; sLon=(lon2-lon1)/(double)nLon;
	--nLat; sLat=(lat2-lat1)/(double)nLat;

	if (argc==12 || argc==14) {
	    ReadImageFile_RGB(argv[argc-3],&iLon,&iLat,&BPS,&R,&G,&B,NULL);
	    fLon=(double)(iLon-1)/(double)nLon;
	    fLat=(double)(iLat-1)/(double)nLat;
	    factor=31.0/(pow(2.0,(double)BPS)-1.0);
	    color=1;
	}
	depth=(-atof(argv[argc-2]));

	Open(argv[argc-1],
	     2*(unsigned)nLon*(unsigned)nLat+6*(unsigned)(nLon+nLat));

	lon=(lon1+lon2)/2.0;
	lat=(lat1+lat2)/2.0;
	XYZ(lon,lat,0.0,&xO,&yO,&zO);
	hx= -sin(lon)	      ; hy=  cos(lon)	      ; hz=     0.0;
	vx= -cos(lon)*sin(lat); vy= -sin(lon)*sin(lat); vz=cos(lat);
	dx=  cos(lon)*cos(lat); dy=  sin(lon)*cos(lat); dz=sin(lat);
	HVD(lon,lat,depth,&hO,&vO,&dO);

#define F1_LON(iLon)	(int)floor(fLon*(double)(iLon*3+2)/3.0+0.5)
#define F1_LAT(iLat)	(int)floor(fLat*(double)(iLat*3+1)/3.0+0.5)
#define F2_LON(iLon)	(int)floor(fLon*(double)(iLon*3+1)/3.0+0.5)
#define F2_LAT(iLat)	(int)floor(fLat*(double)(iLat*3+2)/3.0+0.5)
/*
#define F1_LON(iLon)	(int)floor(fLon*(double)(iLon*3+1)/3.0+0.5)
#define F1_LAT(iLat)	(int)floor(fLat*(double)(iLat*3+1)/3.0+0.5)
#define F2_LON(iLon)	(int)floor(fLon*(double)(iLon*3+2)/3.0+0.5)
#define F2_LAT(iLat)	(int)floor(fLat*(double)(iLat*3+2)/3.0+0.5)
*/
#define FLAG(cLon,cLat)	\
	32768|((int)floor(factor*(double)R[cLat][cLon]+0.5)<<10) \
	     |((int)floor(factor*(double)G[cLat][cLon]+0.5)<<5) \
	     | (int)floor(factor*(double)B[cLat][cLon]+0.5)

#define SQUARE(f1,f2)	Store(h2,v2,d2,h1,v1,d1,h4,v4,d4,f1); \
			Store(h3,v3,d3,h4,v4,d4,h1,v1,d1,f2)
/*
#define SQUARE(f1,f2)	Store(h2,v2,d2,h1,v1,d1,h3,v3,d3,f1); \
			Store(h3,v3,d3,h4,v4,d4,h2,v2,d2,f2)
*/
#define WALL()	SQUARE(0,0); Store(h4,v4,d4,h3,v3,d3,hO,vO,dO,0)

	for (lat=lat2, iLat=0; iLat<nLat; iLat++, lat-=sLat)
	for (lon=lon1, iLon=0; iLon<nLon; iLon++, lon+=sLon) {
	    if (dem!=NULL) {
		e1=base+step*(double)dem[iLat  ][iLon  ];
		e2=base+step*(double)dem[iLat  ][iLon+1];
		e3=base+step*(double)dem[iLat+1][iLon  ];
		e4=base+step*(double)dem[iLat+1][iLon+1];
	    }
	    if (color) {
		cLon=F1_LON(iLon);
		cLat=F1_LAT(iLat); f1=FLAG(cLon,cLat);
		cLon=F2_LON(iLon);
		cLat=F2_LAT(iLat); f2=FLAG(cLon,cLat);
	    }
	    HVD(lon	,lat	 ,e1,&h1,&v1,&d1);
	    HVD(lon+sLon,lat	 ,e2,&h2,&v2,&d2);
	    HVD(lon	,lat-sLat,e3,&h3,&v3,&d3);
	    HVD(lon+sLon,lat-sLat,e4,&h4,&v4,&d4); SQUARE(f1,f2);
	}
	for (lat=lat2, iLat=0; iLat<nLat; iLat++, lat-=sLat) {
	    if (dem!=NULL) {
		e1=base+step*(double)dem[iLat  ][0];
		e2=base+step*(double)dem[iLat+1][0];
	    }
	    HVD(lon1,lat     ,e1   ,&h1,&v1,&d1);
	    HVD(lon1,lat-sLat,e2   ,&h2,&v2,&d2);
	    HVD(lon1,lat     ,depth,&h3,&v3,&d3);
	    HVD(lon1,lat-sLat,depth,&h4,&v4,&d4); WALL();
	}
	for (lon=lon1, iLon=0; iLon<nLon; iLon++, lon+=sLon) {
	    if (dem!=NULL) {
		e1=base+step*(double)dem[nLat][iLon  ];
		e2=base+step*(double)dem[nLat][iLon+1];
	    }
	    HVD(lon	,lat1,e1   ,&h1,&v1,&d1);
	    HVD(lon+sLon,lat1,e2   ,&h2,&v2,&d2);
	    HVD(lon	,lat1,depth,&h3,&v3,&d3);
	    HVD(lon+sLon,lat1,depth,&h4,&v4,&d4); WALL();
	}
	for (lat=lat1, iLat=nLat; iLat>0; iLat--, lat+=sLat) {
	    if (dem!=NULL) {
		e1=base+step*(double)dem[iLat  ][nLon];
		e2=base+step*(double)dem[iLat-1][nLon];
	    }
	    HVD(lon2,lat     ,e1   ,&h1,&v1,&d1);
	    HVD(lon2,lat+sLat,e2   ,&h2,&v2,&d2);
	    HVD(lon2,lat     ,depth,&h3,&v3,&d3);
	    HVD(lon2,lat+sLat,depth,&h4,&v4,&d4); WALL();
	}
	for (lon=lon2, iLon=nLon; iLon>0; iLon--, lon-=sLon) {
	    if (dem!=NULL) {
		e1=base+step*(double)dem[0][iLon  ];
		e2=base+step*(double)dem[0][iLon-1];
	    }
	    HVD(lon	,lat2,e1   ,&h1,&v1,&d1);
	    HVD(lon-sLon,lat2,e2   ,&h2,&v2,&d2);
	    HVD(lon	,lat2,depth,&h3,&v3,&d3);
	    HVD(lon-sLon,lat2,depth,&h4,&v4,&d4); WALL();
	}
	Close();

	return 0;
}
