
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "rif_rgb.h"

extern void	Open(char *path,unsigned vertices,unsigned faces),
		StoreVertex(double x,double y,double z),
		StoreFace(int r,int g,int b,unsigned i,unsigned j,unsigned k),
		Close();

static void	Error(char *msg)
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static double	A2,B2;

static void	XYZ(double lon,double lat,double e,
		    double *x,double *y,double *z)
{
	double	c=cos(lat),s=sin(lat),ab=sqrt(A2*c*c+B2*s*s),xy=(A2/ab+e)*c;

	*x=xy*cos(lon); *y=xy*sin(lon); *z=(B2/ab+e)*s;
}

static double	depth,xO,yO,zO,hx,hy,hz,vx,vy,vz,dx,dy,dz;

static void	HVD(double lon,double lat,double e)
{
	double	x,y,z;

	XYZ(lon,lat,e,&x,&y,&z); x-=xO;
				 y-=yO;
				 z-=zO; StoreVertex(hx*x+hy*y+hz*z,
						    vx*x+vy*y+vz*z,
						    dx*x+dy*y+dz*z);
	XYZ(lon,lat,depth,&x,&y,&z); x-=xO;
				     y-=yO;
				     z-=zO; StoreVertex(hx*x+hy*y+hz*z,
							vx*x+vy*y+vz*z,
							dx*x+dy*y+dz*z);
}

int	main(int argc,char **argv)
{
	double		lon1,lon2,lat1,lat2,base,step,sLon,hLon,sLat,hLat,
			fLon,fLat,factor,lon,lat;
	int		nLon,nLat,iLon,iLat,BPS,r,g,b,cLon,cLat,cR,cG,cB;
	Cell		**dem=NULL,**R=NULL,**G=NULL,**B=NULL;
	unsigned	N,v1,v3,W,v2,v4;

	if (argc<14 || argc>17)
	    Error("usage : ssm_sf_zcp A B lon1 lat1 lon2 lat2\n"
		  "                   (nLon nLat)|(DEM base step ratio)\n"
		  "                   {color}\n"
		  "                   depth R G B ZCP");

	A2=atof(argv[1])*atof(argv[1]);
	B2=atof(argv[2])*atof(argv[2]);

	if ((lon1=atof(argv[3])*M_PI/180.0)>=
	    (lon2=atof(argv[5])*M_PI/180.0) ||
	    (lat1=atof(argv[4])*M_PI/180.0)>=
	    (lat2=atof(argv[6])*M_PI/180.0)) Error("bad map area.");

	if (argc==14 || argc==15) {
	    nLon=atoi(argv[7]);
	    nLat=atoi(argv[8]);
	}
	else {
	    ReadImageFile(argv[7],&nLon,&nLat,NULL,&dem,NULL);
	    base=atof(argv[8])*atof(argv[10]);
	    step=atof(argv[9])*atof(argv[10]);
	}
	if (nLon<=1 || nLat<=1) Error("bad grid number.");

	--nLon; sLon=(lon2-lon1)/(double)nLon; hLon=sLon/2.0;
	--nLat; sLat=(lat2-lat1)/(double)nLat; hLat=sLat/2.0;

	if (argc==15 || argc==17) {
	    ReadImageFile_RGB(argv[argc-6],&iLon,&iLat,&BPS,&R,&G,&B,NULL);
	    fLon=(double)(iLon-1)/(double)nLon;
	    fLat=(double)(iLat-1)/(double)nLat;
	    factor=255.0/(pow(2.0,(double)BPS)-1.0);
	}
	depth=(-atof(argv[argc-5]));

	if ((r=atoi(argv[argc-4]))<0 || r>255 ||
	    (g=atoi(argv[argc-3]))<0 || g>255 ||
	    (b=atoi(argv[argc-2]))<0 || b>255) Error("bad wall color.");

	N=(unsigned)(nLon+2)*(unsigned)(nLat+2);
	Open(argv[argc-1],N*2,(N-1)*4);

	lon=(lon1+lon2)/2.0;
	lat=(lat1+lat2)/2.0;
	XYZ(lon,lat,0.0,&xO,&yO,&zO);
	hx= -sin(lon)	      ; hy=  cos(lon)	      ; hz=     0.0;
	vx= -cos(lon)*sin(lat); vy= -sin(lon)*sin(lat); vz=cos(lat);
	dx=  cos(lon)*cos(lat); dy=  sin(lon)*cos(lat); dz=sin(lat);

	if (dem==NULL) {
	    HVD(lon1,lat2,0.0);
	    for (lon=lon1+hLon, iLon=0; iLon<nLon; iLon++, lon+=sLon)
		HVD(lon,lat2,0.0);
	    HVD(lon2,lat2,0.0);
	    for (lat=lat2-hLat, iLat=0; iLat<nLat; iLat++, lat-=sLat) {
		HVD(lon1,lat,0.0);
		for (lon=lon1+hLon, iLon=0; iLon<nLon; iLon++, lon+=sLon)
		    HVD(lon,lat,0.0);
		HVD(lon2,lat,0.0);
	    }
	    HVD(lon1,lat1,0.0);
	    for (lon=lon1+hLon, iLon=0; iLon<nLon; iLon++, lon+=sLon)
		HVD(lon,lat1,0.0);
	    HVD(lon2,lat1,0.0);
	}
	else {
	    HVD(lon1,lat2,base+step*(double)dem[0][0]);
	    for (lon=lon1+hLon, iLon=0; iLon<nLon; iLon++, lon+=sLon)
		HVD(lon,lat2,base+step*((double)dem[0][iLon  ]+
					(double)dem[0][iLon+1])/2.0);
	    HVD(lon2,lat2,base+step*(double)dem[0][nLon]);
	    for (lat=lat2-hLat, iLat=0; iLat<nLat; iLat++, lat-=sLat) {
		HVD(lon1,lat,base+step*((double)dem[iLat  ][0]+
					(double)dem[iLat+1][0])/2.0);
		for (lon=lon1+hLon, iLon=0; iLon<nLon; iLon++, lon+=sLon)
		    HVD(lon,lat,base+step*((double)dem[iLat  ][iLon  ]+
					   (double)dem[iLat  ][iLon+1]+
					   (double)dem[iLat+1][iLon  ]+
					   (double)dem[iLat+1][iLon+1])/4.0);
		HVD(lon2,lat,base+step*((double)dem[iLat  ][nLon]+
					(double)dem[iLat+1][nLon])/2.0);
	    }
	    HVD(lon1,lat1,base+step*(double)dem[nLat][0]);
	    for (lon=lon1+hLon, iLon=0; iLon<nLon; iLon++, lon+=sLon)
		HVD(lon,lat1,base+step*((double)dem[nLat][iLon  ]+
					(double)dem[nLat][iLon+1])/2.0);
	    HVD(lon2,lat1,base+step*(double)dem[nLat][nLon]);
	}
	v1=0; v3=W=(nLon+2)*2;
	if (R==NULL || G==NULL || B==NULL)
	    for (iLat=0; iLat<=nLat; iLat++, v1+=2, v3+=2)
	    for (iLon=0; iLon<=nLon; iLon++, v1=v2, v3=v4) {
		v2=v1+2; v4=v3+2;
		StoreFace(r,g,b,v2,v1,v4); StoreFace(r,g,b,v4+1,v1+1,v2+1);
		StoreFace(r,g,b,v3,v4,v1); StoreFace(r,g,b,v1+1,v4+1,v3+1);
	    }
	else
	    for (iLat=0; iLat<=nLat; iLat++, v1+=2, v3+=2)
	    for (iLon=0; iLon<=nLon; iLon++, v1=v2, v3=v4) {
		v2=v1+2; v4=v3+2;
		cLon=(int)floor(fLon*(double)iLon+0.5);
		cLat=(int)floor(fLat*(double)iLat+0.5);
		cR=(int)floor(factor*(double)R[cLat][cLon]+0.5);
		cG=(int)floor(factor*(double)G[cLat][cLon]+0.5);
		cB=(int)floor(factor*(double)B[cLat][cLon]+0.5);
		StoreFace(cR,cG,cB,v2,v1,v4); StoreFace(r,g,b,v4+1,v1+1,v2+1);
		StoreFace(cR,cG,cB,v3,v4,v1); StoreFace(r,g,b,v1+1,v4+1,v3+1);
	    }

	for (iLon=0; iLon<=nLon; iLon++, v1=v2) {
	    v2=v1+2; StoreFace(r,g,b,v2,v1,v2+1); StoreFace(r,g,b,v1+1,v2+1,v1);
	}
	for (iLat=0; iLat<=nLat; iLat++, v1=v2) {
	    v2=v1-W; StoreFace(r,g,b,v2,v1,v2+1); StoreFace(r,g,b,v1+1,v2+1,v1);
	}
	for (iLon=nLon; iLon>=0; iLon--, v1=v2) {
	    v2=v1-2; StoreFace(r,g,b,v2,v1,v2+1); StoreFace(r,g,b,v1+1,v2+1,v1);
	}
	for (iLat=nLat; iLat>=0; iLat--, v1=v2) {
	    v2=v1+W; StoreFace(r,g,b,v2,v1,v2+1); StoreFace(r,g,b,v1+1,v2+1,v1);
	}
	Close();

	return 0;
}
