
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "rif_rgb.h"

#define SSM_LON	"SSM_LON"
#define SSM_LAT	"SSM_LAT"

#ifndef	VERTEX_FORMAT
#define VERTEX_FORMAT	"%lf"
#endif

static void	Error(char *msg)
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static double	A2,B2;

static void	Spheroid(double lon,double lat,double e,
			 double *x,double *y,double *z)
{
	double	c=cos(lat),s=sin(lat),ab=sqrt(A2*c*c+B2*s*s),xy=(A2/ab+e)*c;

	*x=xy*cos(lon); *y=xy*sin(lon); *z=(B2/ab+e)*s;
}

static double	depth,xO,yO,zO,hx,hy,hz,vx,vy,vz,dx,dy,dz;

static void	Vertex(double lon,double lat,double e)
{
	double	x,y,z;

	Spheroid(lon,lat,e,&x,&y,&z); x-=xO; y-=yO; z-=zO;
	(void)printf(VERTEX_FORMAT "\t" VERTEX_FORMAT "\t" VERTEX_FORMAT "\n",
		     hx*x+hy*y+hz*z,vx*x+vy*y+vz*z,dx*x+dy*y+dz*z);

	Spheroid(lon,lat,depth,&x,&y,&z); x-=xO; y-=yO; z-=zO;
	(void)printf(VERTEX_FORMAT "\t" VERTEX_FORMAT "\t" VERTEX_FORMAT "\n",
		     hx*x+hy*y+hz*z,vx*x+vy*y+vz*z,dx*x+dy*y+dz*z);
}

#define ULL	unsigned long long

static void	Face_Colorless(ULL v1,ULL v2,ULL v3,ULL v4)
{
	(void)printf("%llu\t%llu\t%llu\n",v1,v4,v2);
	(void)printf("%llu\t%llu\t%llu\n",v1,v3,v4);
}

static void	Face(ULL v1,ULL v2,ULL v3,ULL v4,int r,int g,int b)
{
	(void)printf("%llu\t%llu\t%llu\t%d\t%d\t%d\n",v1,v4,v2,r,g,b);
	(void)printf("%llu\t%llu\t%llu\t%d\t%d\t%d\n",v1,v3,v4,r,g,b);
}

#define INT(d)	(int)floor(d+0.5)

int	main(int argc,char **argv)
{
	double	lon1,lon2,lat1,lat2,base,step,sLon,sLat,fLon,fLat,fC,lon,lat;
	int	nLon,nLat,iLon,iLat,BPS,x,y;
	Cell	**dem=NULL,**R=NULL,**G=NULL,**B=NULL;
	char	*env;
	ULL	w,v1,v2;

	if (argc<10 || argc>13)
	    Error("usage : ssm_vf A B lon1 lat1 lon2 lat2\n"
		  "               (nLon nLat)|(DEM base step ratio)\n"
		  "               {color}\n"
		  "               depth");

	A2=atof(argv[1])*atof(argv[1]);
	B2=atof(argv[2])*atof(argv[2]);

	if ((lon1=atof(argv[3])*M_PI/180.0)>=
	    (lon2=atof(argv[5])*M_PI/180.0) ||
	    (lat1=atof(argv[4])*M_PI/180.0)>=
	    (lat2=atof(argv[6])*M_PI/180.0)) Error("bad map area.");

	if (argc==10 || argc==11) {
	    nLon=atoi(argv[7]);
	    nLat=atoi(argv[8]);
	}
	else {
	    ReadImageFile(argv[7],&nLon,&nLat,NULL,&dem,NULL);
	    base=atof(argv[8])*atof(argv[10]);
	    step=atof(argv[9])*atof(argv[10]);
	}
	if (nLon<=1 || nLat<=1) Error("bad grid number.");

	--nLon; sLon=(lon2-lon1)/(double)nLon;
	--nLat; sLat=(lat2-lat1)/(double)nLat;

	if (argc==11 || argc==13) {
	    ReadImageFile_RGB(argv[argc-2],&iLon,&iLat,&BPS,&R,&G,&B,NULL);
	    fLon=(double)(iLon-1)/(double)nLon;
	    fLat=(double)(iLat-1)/(double)nLat;
	    fC=255.0/(pow(2.0,(double)BPS)-1.0);
	}
	depth=(-atof(argv[argc-1]));

	lon=((env=getenv(SSM_LON))!=NULL)?atof(env)*M_PI/180.0:(lon1+lon2)/2.0;
	lat=((env=getenv(SSM_LAT))!=NULL)?atof(env)*M_PI/180.0:(lat1+lat2)/2.0;

	Spheroid(lon,lat,0.0,&xO,&yO,&zO);

	hx= -sin(lon)	      ; hy=  cos(lon)	      ; hz=     0.0;
	vx= -cos(lon)*sin(lat); vy= -sin(lon)*sin(lat); vz=cos(lat);
	dx=  cos(lon)*cos(lat); dy=  sin(lon)*cos(lat); dz=sin(lat);

	w=((ULL)nLon+1)*((ULL)nLat+1);
	(void)printf("%llu\t%llu\n",w*2,(w-1)*4);

	if (dem==NULL)
	    for (lat=lat2, iLat=0; iLat<=nLat; iLat++, lat-=sLat)
	    for (lon=lon1, iLon=0; iLon<=nLon; iLon++, lon+=sLon)
		Vertex(lon,lat,0.0);
	else
	    for (lat=lat2, iLat=0; iLat<=nLat; iLat++, lat-=sLat)
	    for (lon=lon1, iLon=0; iLon<=nLon; iLon++, lon+=sLon)
		Vertex(lon,lat,base+step*(double)dem[iLat][iLon]);

	v1=0; v2=((ULL)nLon+1)*2;
	if (R==NULL || G==NULL || B==NULL)
	    for (iLat=0; iLat<nLat; iLat++, v1+=2, v2+=2)
	    for (iLon=0; iLon<nLon; iLon++) {
		Face_Colorless(v1,v1+2,v2,v2+2); ++v1; ++v2;
		Face_Colorless(v1+2,v1,v2+2,v2); ++v1; ++v2;
	    }
	else
	    for (iLat=0; iLat<nLat; iLat++, v1+=2, v2+=2)
	    for (iLon=0; iLon<nLon; iLon++) {
		x=INT(fLon*(double)iLon);
		y=INT(fLat*(double)iLat);
		Face(v1,v1+2,v2,v2+2,INT(fC*(double)R[y][x]),
				     INT(fC*(double)G[y][x]),
				     INT(fC*(double)B[y][x])); ++v1; ++v2;
		Face_Colorless(v1+2,v1,v2+2,v2); ++v1; ++v2;
	    }

	v1=0; v2=((ULL)nLon+1)*(ULL)nLat*2;
	for (iLon=0; iLon<nLon; iLon++) {
	    Face_Colorless(v1+2,v1,v1+3,v1+1); v1+=2;
	    Face_Colorless(v2,v2+2,v2+1,v2+3); v2+=2;
	}
	v1=0; v2=(ULL)nLon*2;
	for (iLat=0; iLat<nLat; iLat++) {
	    w=v1+((ULL)nLon+1)*2; Face_Colorless(v1,w,v1+1,w+1); v1=w;
	    w=v2+((ULL)nLon+1)*2; Face_Colorless(w,v2,w+1,v2+1); v2=w;
	}
	return 0;
}
