
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "rif_rgb.h"

extern void	Open(),StoreVertex(),StoreFace(),Close();

static void	Error(msg)
char		*msg;
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static double	A2,B2;

static void	XYZ(lon,lat,e,x,y,z)
double		lon,lat,e,*x,*y,*z;
{
	double	c=cos(lat),s=sin(lat),ab=sqrt(A2*c*c+B2*s*s),xy=(A2/ab+e)*c;

	*x=xy*cos(lon); *y=xy*sin(lon); *z=(B2/ab+e)*s;
}

static double	xO,yO,zO,hx,hy,hz,vx,vy,vz,dx,dy,dz;

static void	HVD(lon,lat,e)
double		lon,lat,e;
{
	double	x,y,z;

	XYZ(lon,lat,e,&x,&y,&z); x-=xO;
				 y-=yO;
				 z-=zO; StoreVertex(hx*x+hy*y+hz*z,
						    vx*x+vy*y+vz*z,
						    dx*x+dy*y+dz*z);
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	double		lon1,lon2,lat1,lat2,base,step,sLon,hLon,sLat,hLat,
			fLon,fLat,factor,depth,lon0,lat0;
	int		nLon,nLat,iLon,iLat,BPS,r,g,b,cLon,cLat,cR,cG,cB;
	Cell		**dem=NULL,**R=NULL,**G=NULL,**B=NULL;
	unsigned	N,W,v1,v2,v3,v4,V;

	if (argc<14 || argc>17)
	    Error("usage : ssm_zcp A B lon1 lat1 lon2 lat2\n"
		  "                (nLon nLat)|(DEM base step ratio)\n"
		  "                {color}\n"
		  "                depth R G B ZCP");

	A2=atof(argv[1])*atof(argv[1]);
	B2=atof(argv[2])*atof(argv[2]);

	if ((lon1=atof(argv[3])*M_PI/180.0)>=
	    (lon2=atof(argv[5])*M_PI/180.0) ||
	    (lat1=atof(argv[4])*M_PI/180.0)>=
	    (lat2=atof(argv[6])*M_PI/180.0)) Error("bad map area.");

	if (argc==14 || argc==15) {
	    nLon=atoi(argv[7]);
	    nLat=atoi(argv[8]);
	}
	else {
	    ReadImageFile(argv[7],&nLon,&nLat,NULL,&dem,NULL);
	    base=atof(argv[8])*atof(argv[10]);
	    step=atof(argv[9])*atof(argv[10]);
	}
	if (nLon<=1 || nLat<=1) Error("bad grid number.");

	--nLon; sLon=(lon2-lon1)/(double)nLon; hLon=sLon/2.0;
	--nLat; sLat=(lat2-lat1)/(double)nLat; hLat=sLat/2.0;

	if (argc==15 || argc==17) {
	    ReadImageFile_RGB(argv[argc-6],&iLon,&iLat,&BPS,&R,&G,&B,NULL);
	    fLon=(double)(iLon-1)/(double)nLon;
	    fLat=(double)(iLat-1)/(double)nLat;
	    factor=255.0/(pow(2.0,(double)BPS)-1.0);
	}
	depth=(-atof(argv[argc-5]));

	if ((r=atoi(argv[argc-4]))<0 || r>255 ||
	    (g=atoi(argv[argc-3]))<0 || g>255 ||
	    (b=atoi(argv[argc-2]))<0 || b>255) Error("bad wall color.");

	N=(unsigned)(nLon+2)*(unsigned)(nLat+2)+(unsigned)(nLon+nLat+2)*2;

	Open(argv[argc-1],N+1,(N-1)*2);

	lon0=(lon1+lon2)/2.0;
	lat0=(lat1+lat2)/2.0;
	XYZ(lon0,lat0,0.0,&xO,&yO,&zO);
	hx= -sin(lon0)		; hy=  cos(lon0)	  ; hz=      0.0;
	vx= -cos(lon0)*sin(lat0); vy= -sin(lon0)*sin(lat0); vz=cos(lat0);
	dx=  cos(lon0)*cos(lat0); dy=  sin(lon0)*cos(lat0); dz=sin(lat0);

#define LON	lon1+hLon+sLon*(double)iLon
#define LAT	lat2-hLat-sLat*(double)iLat

	if (dem==NULL) {
	    HVD(lon1,lat2,0.0);
	    for (iLon=0; iLon<nLon; iLon++) HVD(LON,lat2,0.0);
	    HVD(lon2,lat2,0.0);
	    for (iLat=0; iLat<nLat; iLat++) {
		HVD(lon1,LAT,0.0);
		for (iLon=0; iLon<nLon; iLon++) HVD(LON,LAT,0.0);
		HVD(lon2,LAT,0.0);
	    }
	    HVD(lon1,lat1,0.0);
	    for (iLon=0; iLon<nLon; iLon++) HVD(LON,lat1,0.0);
	    HVD(lon2,lat1,0.0);
	}
	else {
	    HVD(lon1,lat2,base+step*(double)dem[0][0]);
	    for (iLon=0; iLon<nLon; iLon++)
		HVD(LON,lat2,base+step*((double)dem[0][iLon  ]+
					(double)dem[0][iLon+1])/2.0);
	    HVD(lon2,lat2,base+step*(double)dem[0][nLon]);
	    for (iLat=0; iLat<nLat; iLat++) {
		HVD(lon1,LAT,base+step*((double)dem[iLat  ][0]+
					(double)dem[iLat+1][0])/2.0);
		for (iLon=0; iLon<nLon; iLon++)
		    HVD(LON,LAT,base+step*((double)dem[iLat  ][iLon  ]+
					   (double)dem[iLat  ][iLon+1]+
					   (double)dem[iLat+1][iLon  ]+
					   (double)dem[iLat+1][iLon+1])/4.0);
		HVD(lon2,LAT,base+step*((double)dem[iLat  ][nLon]+
					(double)dem[iLat+1][nLon])/2.0);
	    }
	    HVD(lon1,lat1,base+step*(double)dem[nLat][0]);
	    for (iLon=0; iLon<nLon; iLon++)
		HVD(LON,lat1,base+step*((double)dem[nLat][iLon  ]+
					(double)dem[nLat][iLon+1])/2.0);
	    HVD(lon2,lat1,base+step*(double)dem[nLat][nLon]);
	}
	HVD(lon1,lat2,depth);
	for (iLon=0; iLon<nLon; iLon++) HVD(LON,lat2,depth);
	HVD(lon2,lat2,depth);
	for (iLat=0; iLat<nLat; iLat++) HVD(lon2,LAT,depth);
	HVD(lon2,lat1,depth);
	for (iLon=nLon-1; iLon>=0; iLon--) HVD(LON,lat1,depth);
	HVD(lon1,lat1,depth);
	for (iLat=nLat-1; iLat>=0; iLat--) HVD(lon1,LAT,depth);
	HVD(lon0,lat0,depth);

	W=nLon+2;

	v1=0; v2=1; v3=W; v4=W+1;
	if (R==NULL || G==NULL || B==NULL)
	    for (iLat=0; iLat<=nLat; iLat++, v1++, v2++, v3++, v4++)
	    for (iLon=0; iLon<=nLon; iLon++, v1++, v2++, v3++, v4++) {
		StoreFace(r,g,b,v2,v1,v4);
		StoreFace(r,g,b,v3,v4,v1);
	    }
	else
	    for (iLat=0; iLat<=nLat; iLat++, v1++, v2++, v3++, v4++)
	    for (iLon=0; iLon<=nLon; iLon++, v1++, v2++, v3++, v4++) {
		cLon=(int)floor(fLon*(double)iLon+0.5);
		cLat=(int)floor(fLat*(double)iLat+0.5);
		cR=(int)floor(factor*(double)R[cLat][cLon]+0.5);
		cG=(int)floor(factor*(double)G[cLat][cLon]+0.5);
		cB=(int)floor(factor*(double)B[cLat][cLon]+0.5);
		StoreFace(cR,cG,cB,v2,v1,v4);
		StoreFace(cR,cG,cB,v3,v4,v1);
	    }

	V=v1=v3; v3=0;
	for (iLon=0; iLon<=nLon; iLon++) {
	    v2=v1+1; v4=v3+1; StoreFace(r,g,b,v3,v4,v1);
			      StoreFace(r,g,b,v2,v1,v4);
			      StoreFace(r,g,b,v1,v2,N ); v1=v2; v3=v4;
	}
	for (iLat=0; iLat<=nLat; iLat++) {
	    v2=v1+1; v4=v3+W; StoreFace(r,g,b,v3,v4,v1);
			      StoreFace(r,g,b,v2,v1,v4);
			      StoreFace(r,g,b,v1,v2,N ); v1=v2; v3=v4;
	}
	for (iLon=nLon; iLon>=0; iLon--) {
	    v2=v1+1; v4=v3-1; StoreFace(r,g,b,v3,v4,v1);
			      StoreFace(r,g,b,v2,v1,v4);
			      StoreFace(r,g,b,v1,v2,N ); v1=v2; v3=v4;
	}
	for (iLat=nLat; iLat> 0; iLat--) {
	    v2=v1+1; v4=v3-W; StoreFace(r,g,b,v3,v4,v1);
			      StoreFace(r,g,b,v2,v1,v4);
			      StoreFace(r,g,b,v1,v2,N ); v1=v2; v3=v4;
	}
			      StoreFace(r,g,b,v3,0 ,v1);
			      StoreFace(r,g,b,V ,v1,0 );
			      StoreFace(r,g,b,v1,V ,N );

	Close();

	return 0;
}
